#include <Adafruit_NeoPixel.h>
#include <math.h>

// ─────────── Pin Definitions ───────────
#define LDR_LEFT    A0
#define LDR_MID     A1
#define LDR_RIGHT   A2

#define IN1 2
#define IN2 3
#define IN3 4
#define IN4 5
#define ENA 9
#define ENB 10

#define LED_PIN     6
#define NUMPIXELS   144

#define BUTTON_PIN  7

// ─────────── Constants ───────────
const int TOL       = 30;
const int OFFSET0   = 20;
const int OFFSET1   = 5;
const int OFFSET2   = 90;

const unsigned long CALIBRATION_DURATION    = 3000;
const unsigned long WW_BLINK_DURATION       = 2000;
const unsigned long B1_3_BLINK_DURATION     = 3000;
const unsigned long B2_3_LONGBLINK_DURATION = 3000;

const unsigned long BUTTON_STATE1_MIN       = 1000;
const unsigned long BUTTON_STATE1_MAX       = 2000;
const unsigned long BUTTON_STATE2_RESET_MIN = 3000;

// ─────────── LED Colors ───────────
// --- State 1: Warm White ---
#define COLOR_WARM_WHITE         strip.Color(255, 139, 39)
#define COLOR_WARM_WHITE_BRIGHT  strip.Color(255, 255, 234)

// --- State 2: Red ---
#define COLOR_RED_DIM            strip.Color(8, 0, 0)
//#define COLOR_RED_SOLID          strip.Color(255, 0, 0)
#define COLOR_RED_BLINK_DIM      strip.Color(100, 0, 0)
#define COLOR_RED_BLINK_BRIGHT   strip.Color(255, 0, 0)

// --- General ---
#define COLOR_OFF                strip.Color(0, 0, 0)

// ─────────── Type Definitions ───────────
enum PrimaryState { STATE_1, STATE_2 };
enum Behavior1 { B1_1, B1_2, B1_2R, B1_3 };
enum Behavior2 { B2_1, B2_2, B2_3 };

// ─────────── Globals ───────────
Adafruit_NeoPixel strip(NUMPIXELS, LED_PIN, NEO_GRB + NEO_KHZ800);

PrimaryState currentState = STATE_2;
Behavior1 behavior1 = B1_1;
Behavior2 behavior2 = B2_1;

volatile int visitCount = 0;
unsigned long ambientLDR = 0;

int THRESH0 = 0;
int THRESH1 = 0;
int THRESH2 = 0;

bool calibrating = false;
unsigned long calibrateStart = 0;
unsigned long calibrateSum = 0;
unsigned long calibrateCount = 0;

unsigned long blinkStart = 0;
bool b1_2r_counted = false;

bool b2_3_longEntered = false;
unsigned long b2_3_longStart = 0;

unsigned long buttonPressTime = 0;
bool buttonWasDown = false;
bool buttonHandled = false;

// ─────────── Setup ───────────
void setup() {
  Serial.begin(9600);
  pinMode(BUTTON_PIN, INPUT_PULLUP);

  pinMode(IN1, OUTPUT);
  pinMode(IN2, OUTPUT);
  pinMode(IN3, OUTPUT);
  pinMode(IN4, OUTPUT);
  pinMode(ENA, OUTPUT);
  pinMode(ENB, OUTPUT);
  digitalWrite(ENA, HIGH);
  digitalWrite(ENB, HIGH);
  stopMotors();

  strip.begin();
  strip.show();

  currentState = STATE_2;
  behavior2 = B2_1;
  startCalibration();
}

// ─────────── Main Loop ───────────
void loop() {
  int ldrL = analogRead(LDR_LEFT);
  int ldrM = analogRead(LDR_MID);
  int ldrR = analogRead(LDR_RIGHT);

  handleButton();

  // ───── State 1 Behaviors ─────
  if (currentState == STATE_1) {
    switch (behavior1) {
      case B1_1:
        for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_WARM_WHITE);
        strip.show();
        stopMotors();
        if (calibrating) {
          processCalibration(ldrL, ldrM, ldrR);
        } else {
          if (ldrL > THRESH0 || ldrM > THRESH0 || ldrR > THRESH0) {
            behavior1 = B1_2;
            blinkStart = millis();
          }
        }
        printStatus(ldrL, ldrM, ldrR, false);
        break;

      case B1_2: {
        if (ldrL <= THRESH0 && ldrM <= THRESH0 && ldrR <= THRESH0) {
          behavior1 = B1_1;
          startCalibration();
          break;
        }
        unsigned long elapsed = millis() - blinkStart;
        if ((elapsed / 200) % 2 == 0) {
          for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_WARM_WHITE);
        } else {
          for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_WARM_WHITE_BRIGHT);
        }
        strip.show();
        stopMotors();
        if (elapsed >= WW_BLINK_DURATION) {
          behavior1 = B1_2R;
          blinkStart = millis();
          b1_2r_counted = false;
        }
        printStatus(ldrL, ldrM, ldrR, false);
        break;
      }

      case B1_2R:
        if (!b1_2r_counted) {
          visitCount++;
          b1_2r_counted = true;
        }
        if ((millis() - blinkStart) / 200 % 2 == 0) {
          for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_BLINK_DIM);
        } else {
          for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_BLINK_BRIGHT);
        }
        strip.show();
        stopMotors();

        if (visitCount >= 3 && (ldrL > THRESH0 || ldrM > THRESH0 || ldrR > THRESH0)) {
          behavior1 = B1_3;
          blinkStart = millis();
        } else if (ldrL <= THRESH0 && ldrM <= THRESH0 && ldrR <= THRESH0) {
          behavior1 = B1_1;
          startCalibration();
        }
        printStatus(ldrL, ldrM, ldrR, false);
        break;

      case B1_3: {
        if (ldrL <= THRESH0 && ldrM <= THRESH0 && ldrR <= THRESH0) {
          behavior1 = B1_1;
          startCalibration();
          break;
        }
        unsigned long elapsed = millis() - blinkStart;
        if ((elapsed / 200) % 2 == 0) {
          for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_BLINK_DIM);
        } else {
          for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_BLINK_BRIGHT);
        }
        strip.show();
        stopMotors();
        if (elapsed >= B1_3_BLINK_DURATION) {
          for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_OFF);
          strip.show();
          visitCount = 0;
          currentState = STATE_2;
          behavior2 = B2_1;
          Serial.println("Switched to State 2 (via Behavior 1.3)");
          startCalibration();
        }
        printStatus(ldrL, ldrM, ldrR, false);
        break;
      }
    }
  // ───── State 2 Behaviors ─────
  } else {
    switch (behavior2) {
      case B2_1:
        for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_OFF);
        strip.show();
        stopMotors();
        if (calibrating) {
          processCalibration(ldrL, ldrM, ldrR);
        } else if (ldrL > THRESH1 || ldrM > THRESH1 || ldrR > THRESH1) {
          behavior2 = B2_2;
        }
        printStatus(ldrL, ldrM, ldrR, false);
        break;

      case B2_2: {
        if (ldrL <= THRESH1 && ldrM <= THRESH1 && ldrR <= THRESH1) {
            behavior2 = B2_1;
            startCalibration();
            break; 
        }

        bool centeredBright = (ldrM > THRESH2) && (ldrM > ldrL) && (ldrM > ldrR) && (abs(ldrL - ldrR) <= TOL);

        if (!centeredBright) {
          // Move toward the light source
          for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_DIM);
          strip.show();
          moveTowardLight(ldrL, ldrM, ldrR);
          printStatus(ldrL, ldrM, ldrR, true);
        } else {
          // Robot is centered: stop, count visit, switch to blink
          stopMotors();
          visitCount++;
          behavior2 = B2_3;
          blinkStart = millis();
          b2_3_longEntered = false;
          printStatus(ldrL, ldrM, ldrR, false);
        }
        break;
      }

      case B2_3:
        if (ldrL <= THRESH1 && ldrM <= THRESH1 && ldrR <= THRESH1) {
          behavior2 = B2_1;
          startCalibration();
          break;
        }
        
        if (visitCount < 3) {
          if ((millis() - blinkStart) / 200 % 2 == 0) {
            for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_BLINK_DIM);
          } else {
            for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_BLINK_BRIGHT);
          }
          strip.show();
          stopMotors();
        } else {
          if (!b2_3_longEntered) {
            b2_3_longEntered = true;
            b2_3_longStart = millis();
          }
          unsigned long lbElapsed = millis() - b2_3_longStart;
          if (lbElapsed < B2_3_LONGBLINK_DURATION) {
            if ((lbElapsed / 200) % 2 == 0) {
              for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_BLINK_DIM);
            } else {
              for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_RED_BLINK_BRIGHT);
            }
            strip.show();
            stopMotors();
          } else {
            for (int i = 0; i < NUMPIXELS; i++) strip.setPixelColor(i, COLOR_OFF);
            strip.show();
            visitCount = 0;
            currentState = STATE_1;
            behavior1 = B1_1;
            Serial.println("Switched to State 1 (via Behavior 2.3 long blink)");
            startCalibration();
          }
        }
        printStatus(ldrL, ldrM, ldrR, false);
        break;
    }
  }
}

// ─────────── Button Handler ───────────
void handleButton() {
  bool buttonDown = (digitalRead(BUTTON_PIN) == LOW);
  unsigned long now = millis();

  if (buttonDown && !buttonWasDown) {
    buttonPressTime = now;
    buttonWasDown = true;
    buttonHandled = false;
  }
  if (!buttonDown && buttonWasDown) {
    unsigned long duration = now - buttonPressTime;
    if (!buttonHandled) {
      // Very short press (<500 ms): go to State 2 Behavior 2.1 (keep visitCount)
      if (duration < 500) {
        currentState = STATE_2;
        behavior2    = B2_1;
        Serial.println("Short press: switched to Behavior 2.1");
        startCalibration();
        buttonHandled = true;

      // 1–2 s press → State 1, reset visits
      } else if (duration >= BUTTON_STATE1_MIN && duration <= BUTTON_STATE1_MAX) {
        visitCount   = 0;
        currentState = STATE_1;
        behavior1    = B1_1;
        Serial.println("Button: switched to State 1");
        startCalibration();
        buttonHandled = true;

      // ≥3 s press → State 2 reset (visits reset)
      } else if (duration >= BUTTON_STATE2_RESET_MIN) {
        visitCount   = 0;
        currentState = STATE_2;
        behavior2    = B2_1;
        Serial.println("Button: switched to State 2");
        startCalibration();
        buttonHandled = true;
      }
    }
    buttonWasDown = false;
  }
}

// ─────────── Calibration ───────────
void startCalibration() {
  calibrating   = true;
  calibrateStart= millis();
  calibrateSum  = 0;
  calibrateCount= 0;
}

void processCalibration(int ldrL, int ldrM, int ldrR) {
  unsigned long now = millis();
  if (now - calibrateStart < CALIBRATION_DURATION) {
    calibrateSum   += ldrL + ldrM + ldrR;
    calibrateCount += 3;
  } else {
    ambientLDR = calibrateSum / calibrateCount;
    THRESH0    = ambientLDR + OFFSET0;
    THRESH1    = ambientLDR + OFFSET1;
    THRESH2    = ambientLDR + OFFSET2;
    calibrating= false;
  }
}

// ─────────── Motor Control & Helpers ───────────
void moveTowardLight(int l, int m, int r) {
  if ((m > l) && (m > r) && (abs(l - r) <= TOL)) {
    goForward();
  } else if (l > r + TOL) {
    turnLeft();
  } else if (r > l + TOL) {
    turnRight();
  } else {
    goForward();
  }
}

void goForward() {
  digitalWrite(IN1, LOW);  digitalWrite(IN2, HIGH);
  digitalWrite(IN3, HIGH); digitalWrite(IN4, LOW);
}

void turnLeft() {
  digitalWrite(IN1, HIGH); digitalWrite(IN2, LOW);
  digitalWrite(IN3, HIGH); digitalWrite(IN4, LOW);
}

void turnRight() {
  digitalWrite(IN1, LOW);  digitalWrite(IN2, HIGH);
  digitalWrite(IN3, LOW);  digitalWrite(IN4, HIGH);
}

void stopMotors() {
  digitalWrite(IN1, LOW); digitalWrite(IN2, LOW);
  digitalWrite(IN3, LOW); digitalWrite(IN4, LOW);
}

// ─────────── Debug Status Printer ───────────
void printStatus(int l, int m, int r, bool motorRunning) {
  Serial.print("State: ");
  Serial.print((currentState == STATE_1) ? "1" : "2");
  Serial.print(" | Behavior: ");
  if (currentState == STATE_1) {
    if (behavior1 == B1_1)      Serial.print("1.1");
    else if (behavior1 == B1_2)  Serial.print("1.2");
    else if (behavior1 == B1_2R) Serial.print("1.2R");
    else                         Serial.print("1.3");
  } else {
    if (behavior2 == B2_1)      Serial.print("2.1");
    else if (behavior2 == B2_2)  Serial.print("2.2");
    else                         Serial.print("2.3");
  }
  Serial.print(" | L: "); Serial.print(l);
  Serial.print("  M: "); Serial.print(m);
  Serial.print("  R: "); Serial.print(r);
  Serial.print("  Amb: "); Serial.print(ambientLDR);
  Serial.print("  | visits: "); Serial.print(visitCount);
  Serial.print("  | motors: ");
  Serial.println(motorRunning ? "running" : "stop");
}
